@@ -14,18 +14,19 @@ import random
 import pytest
 
 from hypothesis import core, find, given, register_random, strategies as st
-from hypothesis.errors import InvalidArgument
+from hypothesis.errors import HypothesisWarning, InvalidArgument
 from hypothesis.internal import entropy
-from hypothesis.internal.compat import PYPY
+from hypothesis.internal.compat import GRAALPY, PYPY
 from hypothesis.internal.entropy import deterministic_PRNG
 
 
-def gc_on_pypy():
+def gc_collect():
     # CPython uses reference counting, so objects (without circular refs)
     # are collected immediately on `del`, breaking weak references.
-    # PyPy doesn't, so we use this function in tests before counting the
+    # Python implementations with other garbage collection strategies may
+    # or may not, so we use this function in tests before counting the
     # surviving references to ensure that they're deterministic.
-    if PYPY:
+    if PYPY or GRAALPY:
         gc.collect()
 
 
@@ -54,15 +55,18 @@ def test_cannot_register_non_Random():
         register_random("not a Random instance")
 
 
+@pytest.mark.filterwarnings(
+    "ignore:It looks like `register_random` was passed an object that could be garbage collected"
+)
 def test_registering_a_Random_is_idempotent():
-    gc_on_pypy()
+    gc_collect()
     n_registered = len(entropy.RANDOMS_TO_MANAGE)
     r = random.Random()
     register_random(r)
     register_random(r)
     assert len(entropy.RANDOMS_TO_MANAGE) == n_registered + 1
     del r
-    gc_on_pypy()
+    gc_collect()
     assert len(entropy.RANDOMS_TO_MANAGE) == n_registered
 
 
@@ -131,7 +135,6 @@ def test_given_does_not_pollute_state():
 
 def test_find_does_not_pollute_state():
     with deterministic_PRNG():
-
         find(st.random_module(), lambda r: True)
         state_a = random.getstate()
         state_a2 = core._hypothesis_global_random.getstate()
@@ -144,8 +147,11 @@ def test_find_does_not_pollute_state():
         assert state_a2 != state_b2
 
 
+@pytest.mark.filterwarnings(
+    "ignore:It looks like `register_random` was passed an object that could be garbage collected"
+)
 def test_evil_prng_registration_nonsense():
-    gc_on_pypy()
+    gc_collect()
     n_registered = len(entropy.RANDOMS_TO_MANAGE)
     r1, r2, r3 = random.Random(1), random.Random(2), random.Random(3)
     s2 = r2.getstate()
@@ -160,7 +166,7 @@ def test_evil_prng_registration_nonsense():
 
     with deterministic_PRNG(0):
         del r1
-        gc_on_pypy()
+        gc_collect()
         assert k not in entropy.RANDOMS_TO_MANAGE, "r1 has been garbage-collected"
         assert len(entropy.RANDOMS_TO_MANAGE) == n_registered + 1
 
@@ -172,3 +178,57 @@ def test_evil_prng_registration_nonsense():
     # Implicit check, no exception was raised in __exit__
     assert r2.getstate() == s2, "reset previously registered random state"
     assert r3.getstate() == s4, "retained state when registered within the context"
+
+
+@pytest.mark.skipif(
+    PYPY, reason="We can't guard against bad no-reference patterns in pypy."
+)
+def test_passing_unreferenced_instance_raises():
+    with pytest.raises(ReferenceError):
+        register_random(random.Random(0))
+
+
+@pytest.mark.skipif(
+    PYPY, reason="We can't guard against bad no-reference patterns in pypy."
+)
+def test_passing_unreferenced_instance_within_function_scope_raises():
+    def f():
+        register_random(random.Random(0))
+
+    with pytest.raises(ReferenceError):
+        f()
+
+
+@pytest.mark.skipif(
+    PYPY, reason="We can't guard against bad no-reference patterns in pypy."
+)
+def test_passing_referenced_instance_within_function_scope_warns():
+    def f():
+        r = random.Random(0)
+        register_random(r)
+
+    with pytest.warns(
+        HypothesisWarning,
+        match="It looks like `register_random` was passed an object that could be"
+        " garbage collected",
+    ):
+        f()
+
+
+@pytest.mark.filterwarnings(
+    "ignore:It looks like `register_random` was passed an object that could be garbage collected"
+)
+@pytest.mark.skipif(
+    PYPY, reason="We can't guard against bad no-reference patterns in pypy."
+)
+def test_register_random_within_nested_function_scope():
+    n_registered = len(entropy.RANDOMS_TO_MANAGE)
+
+    def f():
+        r = random.Random()
+        register_random(r)
+        assert len(entropy.RANDOMS_TO_MANAGE) == n_registered + 1
+
+    f()
+    gc_collect()
+    assert len(entropy.RANDOMS_TO_MANAGE) == n_registered
